package org.opengis.cite.geomatics;

import java.util.Iterator;
import javax.xml.XMLConstants;
import javax.xml.namespace.NamespaceContext;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Provides static methods for manipulating GML elements.
 * 
 */
public class GmlUtils {

    public static final String SRS_NAME = "srsName";
    public static final String GML_NS = "http://www.opengis.net/gml/3.2";

    /**
     * Sets the srsName attribute on all members of a GML geometry collection if
     * it is specified for the collection. If a geometry member already has a
     * srsName attribute it is left as is.
     * 
     * @param geometryNodes
     *            A sequence of GML geometry elements.
     * 
     * @see "ISO 19136, cl. 10.1.3.2: SRSReferenceGroup"
     */
    public static void setSrsNameOnCollectionMembers(Node... geometryNodes) {
        for (Node geomNode : geometryNodes) {
            Element geom = (Element) geomNode;
            String geomType = geom.getLocalName();
            String srsName = geom.getAttribute(SRS_NAME);
            if (!geomType.startsWith("Multi") || srsName.isEmpty()) {
                continue;
            }
            String memberType = geomType.substring(5).toLowerCase();
            // both kinds of properties may appear in the same collection
            String expr = String.format("gml:%sMember/* | gml:%<sMembers/*",
                    memberType);
            XPath xpath = XPathFactory.newInstance().newXPath();
            xpath.setNamespaceContext(new GmlNamespaceContext());
            NodeList members;
            try {
                members = (NodeList) xpath.evaluate(expr, geom,
                        XPathConstants.NODESET);
            } catch (XPathExpressionException xpe) {
                throw new RuntimeException(xpe);
            }
            for (int i = 0; i < members.getLength(); i++) {
                Element member = (Element) members.item(i);
                if (member.getAttribute(SRS_NAME).isEmpty()) {
                    member.setAttribute(SRS_NAME, srsName);
                }
            }
        }
    }

    /**
     * Returns the value of the srsName attribute for the given geometry
     * element. If a geometry element does not explicitly carry the srsName
     * attribute, then it shall be inherited from either:
     * <ol>
     * <li>the nearest ancestor geometry (aggregate) that has the srsName
     * attribute, or</li>
     * <li>the gml:boundedBy/gml:Envelope element in the containing feature
     * instance.</li>
     * </ol>
     * 
     * <p>
     * As a side effect, an implicit CRS reference will be added to the element
     * using the inherited srsName value.
     * </p>
     * 
     * @param geom
     *            An Element representing a GML geometry object.
     * @return A String denoting a CRS reference (an absolute URI value), or an
     *         empty string if no reference was found.
     * 
     * @see "ISO 19136, cl. 9.10, 10.1.3.2"
     */
    public static String findCRSReference(Element geom) {
        String expr = "./ancestor-or-self::*[@srsName][1]/@srsName";
        XPath xpath = XPathFactory.newInstance().newXPath();
        String srsName;
        try {
            srsName = (String) xpath
                    .evaluate(expr, geom, XPathConstants.STRING);
            if (srsName.isEmpty()) {
                // Look for envelope of containing feature
                NamespaceContext nsContext = new NodeNamespaceContext(geom);
                String gmlPrefix = nsContext.getPrefix(GML_NS);
                expr = String
                        .format("./ancestor::*[%s:boundedBy][1]/%1$s:boundedBy/%1$s:Envelope/@srsName",
                                gmlPrefix);
                xpath.setNamespaceContext(nsContext);
                srsName = (String) xpath.evaluate(expr, geom,
                        XPathConstants.STRING);
            }
        } catch (XPathExpressionException xpe) {
            throw new RuntimeException(xpe);
        }
        if (!srsName.isEmpty()) {
            geom.setAttribute("srsName", srsName);
        }
        return srsName;
    }

    /**
     * A NamespaceContext that only recognizes the conventional "gml" namespace
     * prefix.
     */
    static class GmlNamespaceContext implements NamespaceContext {
        @Override
        public String getNamespaceURI(String prefix) {
            String nsName;
            if (prefix.equals("gml")) {
                nsName = GML_NS;
            } else {
                nsName = XMLConstants.NULL_NS_URI;
            }
            return nsName;
        }

        @Override
        public String getPrefix(String namespaceURI) {
            String prefix = null;
            if (namespaceURI.equals(GML_NS)) {
                prefix = "gml";
            }
            return prefix;
        }

        @Override
        public Iterator<String> getPrefixes(String namespaceURI) {
            return null;
        }
    }

    /**
     * A NamespaceContext that provides the in-scope namespace bindings for a
     * given DOM Node.
     */
    static class NodeNamespaceContext implements NamespaceContext {

        private Node sourceNode;

        public NodeNamespaceContext(Node node) {
            sourceNode = node;
        }

        @Override
        public String getNamespaceURI(String prefix) {
            if (prefix.equals(XMLConstants.DEFAULT_NS_PREFIX)) {
                return sourceNode.lookupNamespaceURI(null);
            } else {
                return sourceNode.lookupNamespaceURI(prefix);
            }
        }

        @Override
        public String getPrefix(String namespaceURI) {
            return sourceNode.lookupPrefix(namespaceURI);
        }

        @Override
        public Iterator<String> getPrefixes(String namespaceURI) {
            return null;
        }
    }

}
